/**
 * Use a Higher UI to achieve the goal of reselect;
* */
import React, {Component} from 'react';
import {connect} from 'react-redux';
import invariant from 'fbjs/lib/invariant';

function disassembleParamsAndSelector(mapStateToProps, state) {
    if (!mapStateToProps) return {};
    invariant(typeof mapStateToProps === 'function', "function connetUI\'s first param must be the type of function , null, undefine");
    let returnValue = mapStateToProps(state);
    let stateOnProps = {};
    for (let key in returnValue) {
        if (!returnValue.hasOwnProperty(key)) continue;
        let value = returnValue[key];
        stateOnProps[key] = {};
        if (Array.isArray(value) && (value[value.length -1] instanceof Function)) {
            invariant(value.length > 1, 'function connetUI mapStateToProps\'s selectors list must larger than 1');
            stateOnProps[key].params = value.slice(0, value.length - 1);
            stateOnProps[key].selector = value[value.length -1];
        }else if (value instanceof Function) {
            stateOnProps[key].params = [state];
            stateOnProps[key].selector = value;
        }else {
            stateOnProps[key].params = [state];
            stateOnProps[key].selector = () => value;
        }
    }
    return stateOnProps;
}

function calculateNewChildProps(newSelectorObj, oldSelectorObj, oldProps) {
    let nextProps = {};
    for (let key in newSelectorObj) {
        if (!newSelectorObj.hasOwnProperty(key)) return;
        let newSelectorObjForKey = newSelectorObj[key];
        if (oldSelectorObj && oldProps) {
            let oldSelectorObjForKey = oldSelectorObj[key];
            let paramsIsChange = newSelectorObjForKey.params.some((item, index) => item !== oldSelectorObjForKey.params[index]);
            if (!paramsIsChange) {
                nextProps[key] = oldProps[key];
            }else {
                nextProps[key] = newSelectorObjForKey.selector(...newSelectorObjForKey.params);
            }
        }else {
            nextProps[key] = newSelectorObjForKey.selector(...newSelectorObjForKey.params);
        }
    }
    return nextProps;
}

/**
 * 检查state是否发生了改变, 浅比较
 * @param thisState
 * @param nextState
 * @returns {boolean}
 */
function isStateChange(thisState, nextState) {
    for (let key in nextState) {
        if (thisState[key] !== nextState[key]) {
            return true;
        }
    }
    return false;
}


function connetUI(...args) {
    return function (UI) {
        args = args || [];
        let mapStateToProps = args[0];
        let mapDispatchToProps = args[1];
        let withRef = !!args[2];
        class NewUI extends Component {
            constructor(props) {
                super(props);
                this.paramsAndSelectors = disassembleParamsAndSelector(mapStateToProps, props.$$reduxState);
                this.state = {
                    childProps: calculateNewChildProps(this.paramsAndSelectors)
                };
            }
            _refView = null;
            get inner() {
                return this._refView
            }
            componentWillReceiveProps(nextProps) {
                let paramsAndSelectors = disassembleParamsAndSelector(mapStateToProps, nextProps.$$reduxState);
                let newChildProps = calculateNewChildProps(paramsAndSelectors, this.paramsAndSelectors, this.state.childProps);
                this.setState({
                    childProps: newChildProps
                });
                this.paramsAndSelectors = paramsAndSelectors;
            }

            shouldComponentUpdate(nextProps, nextState) {
                return isStateChange(this.state.childProps, nextState.childProps);
            }
            render() {
                return <UI {...this.props} {...this.state.childProps} ref={view => this._refView = view}/>
            }
        }
        return connect(state => ({$$reduxState: state}), mapDispatchToProps, null, {withRef: withRef})(NewUI);
    }
}


module.exports = {
    connetUI
}